@@ -11,7 +11,7 @@ use AnyEvent::WebSocket::Message;
 use Carp ();
 
 # ABSTRACT: WebSocket connection for AnyEvent
-our $VERSION = '0.34'; # VERSION
+our $VERSION = '0.35'; # VERSION
 
 
 has handle => (
@@ -53,9 +53,18 @@ sub BUILD
 {
   my $self = shift;
   Scalar::Util::weaken $self;
+  
+  my @temp_messages = ();
+  my $are_callbacks_supposed_to_be_ready = 0;
+  
   my $finish = sub {
     my $strong_self = $self; # preserve $self because otherwise $self can be destroyed in the callbacks.
     return if $self->_is_finished;
+    eval
+    {
+      $self->_process_message($_) foreach @temp_messages;
+    };
+    @temp_messages = ();
     $self->_is_finished(1);
     $self->handle->push_shutdown;
     $self->_is_read_open(0);
@@ -77,53 +86,96 @@ sub BUILD
       while(defined(my $body = $frame->next_bytes))
       {
         next if !$self->_is_read_open; # not 'last' but 'next' in order to consume data in $frame buffer.
-        if($frame->is_text || $frame->is_binary)
+        my $message = AnyEvent::WebSocket::Message->new(
+          body   => $body,
+          opcode => $frame->opcode,
+        );
+        if($are_callbacks_supposed_to_be_ready)
         {
-          my $message = AnyEvent::WebSocket::Message->new(
-            body   => $body,
-            opcode => $frame->opcode,
-          );
-          
-          $_->($self, $message) for @{ $self->_next_message_cb };
-          @{ $self->_next_message_cb } = ();
-          $_->($self, $message) for @{ $self->_each_message_cb };
+          $self->_process_message($message);
         }
-        elsif($frame->is_close)
+        else
         {
-          $self->_is_read_open(0);
-          $self->close();
+          push(@temp_messages, $message);
         }
-        elsif($frame->is_ping)
-        {
-          $self->send(AnyEvent::WebSocket::Message->new(opcode => 10, body => $body));
-        }
-      }
+      };
       1; # succeed to parse.
     };
     if(!$success)
     {
-      $self->handle->push_shutdown;
-      $self->_is_write_open(0);
-      $self->_is_read_open(0);
+      $self->_force_shutdown();
     }
   };
 
-  # Delay setting on_read callback. This is necessary to make sure all
-  # received data are handled by each_message and/or next_message
-  # callbacks. If there is some data in rbuf, changing the on_read
-  # callback makes the callback fire, but there is of course no
-  # each_message/next_message callback to receive the message yet.
-  $self->handle->on_read(undef);
+
+  # Message processing (calling _process_message) is delayed by
+  # $are_callbacks_supposed_to_be_ready flag. This is necessary to
+  # make sure all received messages are delivered to each_message and
+  # next_message callbacks. If there is some data in rbuf, changing
+  # the on_read callback makes the callback fire, but there is of
+  # course no each_message/next_message callback to receive the
+  # message yet. So we put messages to @temp_messages for a
+  # while. After the control is returned to the user, who sets up
+  # each_message/next_message callbacks, @temp_messages are processed.
+
+  # An alternative approach would be temporarily disabling on_read by
+  # $self->handle->on_read(undef). However, this can cause a weird
+  # situation in TLS mode, because on_eof can fire even if we don't
+  # have any on_read (
+  # https://metacpan.org/pod/AnyEvent::Handle#I-get-different-callback-invocations-in-TLS-mode-Why-cant-I-pause-reading
+  # )   
+  $self->handle->on_read($read_cb);
   my $idle_w; $idle_w = AE::idle sub {
     undef $idle_w;
     if(defined($self))
     {
-      $read_cb->($self->handle); # make sure to read remaining data in rbuf.
-      $self->handle->on_read($read_cb);
+      my $strong_self = $self;
+      $are_callbacks_supposed_to_be_ready = 1;
+      local $@;
+      my $success = eval
+      {
+        $self->_process_message($_) foreach @temp_messages;
+        1;
+      };
+      @temp_messages = ();
+      if(!$success)
+      {
+        $self->_force_shutdown();
+      }
     }
   };
 }
 
+sub _process_message
+{
+  my ($self, $received_message) = @_;
+  return if !$self->_is_read_open;
+  
+  if($received_message->is_text || $received_message->is_binary)
+  {
+    $_->($self, $received_message) for @{ $self->_next_message_cb };
+    @{ $self->_next_message_cb } = ();
+    $_->($self, $received_message) for @{ $self->_each_message_cb };
+  }
+  elsif($received_message->is_close)
+  {
+    $self->_is_read_open(0);
+    $self->close();
+  }
+  elsif($received_message->is_ping)
+  {
+    $self->send(AnyEvent::WebSocket::Message->new(opcode => 10, body => $received_message->body));
+  }
+}
+
+sub _force_shutdown
+{
+  my ($self) = @_;
+  $self->handle->push_shutdown;
+  $self->_is_write_open(0);
+  $self->_is_read_open(0);
+}
+
 
 sub send
 {
@@ -201,7 +253,7 @@ AnyEvent::WebSocket::Connection - WebSocket connection for AnyEvent
 
 =head1 VERSION
 
-version 0.34
+version 0.35
 
 =head1 SYNOPSIS
 
